/******************************************************************************
*
*	Simple Reliable Communication System over Ethernet _ Single MicroBlaze System 
*
*	Source Code for Handler Functions	
* 
* 	Author: 	Patrick E. Akl  
*           	American University of Beirut
*				August 2004
*			 	
*	For any comments, suggestions or bug discovery, please contact me at the following
*	email address : patrickakl@hotmail.com
*					
*	or contact Christopher-John Comis (University of Toronto)
*   at the following email address:
*					comis@eecg.toronto.edu
*
*******************************************************************************/

               
#include "SingleMB_InitializationFunctions.h"
#include "SingleMB_GlobalVariables.h"
#include "SingleMB_MACAddresses.h"

/*
* Initializes the GPio
*
* Led is an output
*/
static XStatus Initialize_GPio()								
{
	XStatus Status;
    Status = XGpio_Initialize(&Gpio, XPAR_OPB_GPIO_0_DEVICE_ID);
    if (Status != XST_SUCCESS)
    {
        xil_printf("GPIO initialization error !!\n\r");
    }
	else   
		xil_printf("GPIO initialization succeeded !!\n\r");
    /* Set the direction for all signals to be inputs except the LED 
     * outputs */
    XGpio_SetDataDirection(&Gpio, ~LED);

	return Status;
}



/*
*
* Initializes the ACK Frame common fields 
* so that we don't do it every time we send an ACK
* at run time. 
*
* This is to reduce the computational overhead 
* The Ack Frame is contained in the Buffer ACKBuffer[60]
*
* ACK Format:
*
*   DST ADDRESS_SRC ADDRESS_ETHERNET LEN/TYPE FIELD_MESSAGE TYPE_SEQUENCE NUMBERS_MESSAGE LEN_PAYLOAD
*    unknown     automatic           known				known		unknown			known	   known        
*/
static void InitializeACKFrameCommonFields()
{
	Xuint32 loopcounter;
	Xuint8* ACKBuffPointer;
	ACKBuffPointer = (Xuint8 *) AckFrame;

		// Get Passed Destination and Source MAC Address
	ACKBuffPointer += (2 * XEM_MAC_ADDR_SIZE);

		// Write Known ETHERNET LEN / TYPE Field _ 46
	*(Xuint16 *)ACKBuffPointer = ACK_L1_PAYLOAD_SIZE;
	
		// Advance the pointer to the position of Layer 2 Header
	ACKBuffPointer += 2;
	
		// Write Known Message Type _ 1 
	*(Xuint16 *)ACKBuffPointer = ACK_MESSAGE_TYPE;

		// Get Passed SeqNum Field _ Get to Layer2 Len Field
	ACKBuffPointer += 6;

		// Write Known Layer 2 Payload Length Field
	*(Xuint32 *)ACKBuffPointer = ACK_L2_PAYLOAD_SIZE;

		// Advance Pointer to the Layer2 Known ACK Payload
	ACKBuffPointer += 4;
	
		//Writing Known Layer2 ACK Payload
	for(loopcounter = 0 ; loopcounter < ACK_L2_PAYLOAD_SIZE ; loopcounter++)
	{
		*(Xuint8 *)ACKBuffPointer++ = 0xFF;
	}
}


/*
* Setup Interrupt System
*
* EMAC has highest priority			0
* TIMER COUNTER has lowest priority 1
*/
static XStatus SetupInterruptSystem(XEmac *EmacPtr)			/* Setup Interrupt System */
{
    XStatus Result;

    microblaze_enable_interrupts();

    Result = XIntc_Initialize(&InterruptController, INTC_DEVICE_ID); // 0 is DEVICE ID for the Interrupt Controller   
	if (Result != XST_SUCCESS)
    {
        xil_printf("Error in Xintc Initialization !!\n\r");
		return Result;
    }
	else
        xil_printf("Xintc Initialization Succeeded !!\n\r");



 	Result = XIntc_Connect(&InterruptController, INTC_INPUT_TIMER, (XInterruptHandler)XTmrCtr_InterruptHandler, &TimerCounter);
    if (Result != XST_SUCCESS)  
    {
        xil_printf("Error in Xintc Connect to Timer !!\n\r");
		return Result;
    }
	else
        xil_printf("Xintc Connect Succeeded to Timer!!\n\r");


	Result = XIntc_Connect(&InterruptController, INTC_INPUT_EMAC, (XInterruptHandler)XEmac_IntrHandlerFifo, EmacPtr);
    if (Result != XST_SUCCESS)
    {
		return Result;
        xil_printf("Error in Xintc Connect to EMAC!!\n\r");    
    }
	else
        xil_printf("Xintc Connect Succeeded to EMAC!!\n\r");
								// all interrupts will be serviced in order of occurence
	Result = XIntc_SetOptions(&InterruptController, XIN_SVC_ALL_ISRS_OPTION);  
	if (Result != XST_SUCCESS)
    {
		return Result;
        xil_printf("Error in Xintc SetOptions !!\n\r");
    }
	else
        xil_printf("Xintc SetOptions Succeeded !!\n\r");


	Result = XIntc_Start(&InterruptController, XIN_REAL_MODE);  
	if (Result != XST_SUCCESS)
    {
		return Result;
        xil_printf("Error in Xintc Start !!\n\r");
    }
	else
        xil_printf("Xintc Start Succeeded !!\n\r");
	

    XIntc_Enable(&InterruptController, INTC_INPUT_EMAC); // enable the EMAC interrupt
	//INTC_INPUT_TIMER is enabled in the SendReliably function after a send

    return XST_SUCCESS; 
}

/*
* Initialize the Timer Counter
*/
static XStatus InitializeTimerCounter()
{
	XStatus Status;
		// Disabling the counter if some are running with low level funtions
	XTmrCtr_mDisable(0x80010400, TIMER_COUNTER_0);   

		// Disable the input of the interrupt controller
	XIntc_mMasterDisable(0x80010500);  

    /************************* Timer Setup ********************************/ 

    Status = XTmrCtr_Initialize(&TimerCounter, TIMER_COUNTER_0);  
						
    if (Status != XST_SUCCESS)
    {
        xil_printf("Timer counter initialization error\n\r\r");
		return Status;
    }
	
    Status = XTmrCtr_SelfTest(&TimerCounter, TIMER_COUNTER_0);
    if (Status != XST_SUCCESS)
    {
        xil_printf("Timer counter self-test error\n\r");
		return Status;
    }
   
    XTmrCtr_SetHandler(&TimerCounter, TimerCounterHandler, &TimerCounter);

    XTmrCtr_SetOptions(&TimerCounter, TIMER_COUNTER_0, XTC_INT_MODE_OPTION | XTC_AUTO_RELOAD_OPTION);

    XTmrCtr_SetResetValue(&TimerCounter, TIMER_COUNTER_0, 0xF0000000);

    return XST_SUCCESS; 
}

/*
* Initialize the EMAC Module
*/
static XStatus InitializeEmac()
{
	XEmac_Config *ConfigPtr;	
	XStatus Result;
	Xuint32 DeviceId;
	DeviceId = 0;																	
	EmacPtr = & Emac;

    ConfigPtr = XEmac_LookupConfig(DeviceId);        
    ConfigPtr->IpIfDmaConfig = XEM_CFG_NO_DMA;

    Result = XEmac_Initialize(EmacPtr, DeviceId);
    if (Result != XST_SUCCESS)
    {	
		xil_printf("Error in EMAC Initialization !!\n\r");
        return Result;
    }
	else
		xil_printf("Initialization Succeeded !!\n\r");


    Result = XEmac_SelfTest(EmacPtr);
    if (Result != XST_SUCCESS)
    {
		xil_printf("Self Test Error !!\n\r");
        return Result;
    }
	else
		xil_printf("Self Test Succeeded !!\n\r");

    Result = XEmac_SetOptions(EmacPtr, TEST_1_OPTIONS);
    if (Result != XST_SUCCESS)
    {
		xil_printf("Set Options Error !!\n\r");
        return Result;
    }
	else
		xil_printf("Set Options Succeeded !!\n\r");

    Result = XEmac_SetMacAddress(EmacPtr, LocalAddress); 
    if (Result != XST_SUCCESS)
    {
		xil_printf("Set MAC Address Error !!\n\r");
        return Result;
    }
	else
		xil_printf("Set MAC Address Succeeded !!\n\r");

    XEmac_SetFifoSendHandler(EmacPtr, EmacPtr, FifoSendHandler);
    XEmac_SetFifoRecvHandler(EmacPtr, EmacPtr, FifoRecvHandler);
    XEmac_SetErrorHandler(EmacPtr, EmacPtr, ErrorHandler);


	XEmac_ClearStats(EmacPtr);
    Result = XEmac_Start(EmacPtr);
    if (Result != XST_SUCCESS)
    {
		xil_printf("EMAC Start Error !!\n\r");
        return Result;
    }
	else
		xil_printf("EMAC Start Success !!\n\r");

	return XST_SUCCESS;
}




/*
* This function is called from Initialize Parameters (which is the only one called from main)
* and hence there is no need to
* call it from main.
* 
* It initializes the Communication Parameters Structure contents 
*
* It assumes we have 5 boards (check the NUMBER_OF_BOARDS variable in Global Variables)
*/
void Initialize_CommVars()
{
	Xuint8 cntr;

	for(cntr=0 ; cntr < NUMBER_OF_BOARDS ; cntr++)
	{
		CommVars[cntr].SeqNum = 0;
		CommVars[cntr].WaitingFrameNumber = 0;
		CommVars[cntr].Ack_Received = 0;
		CommVars[cntr].ExpectMore = 0;
		CommVars[cntr].RememberSeqNumReceiver = 0;
		CommVars[cntr].NbFragRecv = 0;
		CommVars[cntr].MsgLength = 0;
		CommVars[cntr].FragLength = 0;
	}
}


/*
* This function should be called from main
* It calls all the other functions.
*/
void Initialize_Parameters()					
{	
	XStatus Result;

	Timer_Expired = 1;     	  	     
	reload_flag=0;

	Initialize_CommVars();

	InitializeACKFrameCommonFields();

	Result = InitializeEmac();
	if (Result != XST_SUCCESS)
    {
		xil_printf("EMAC Initialization Error !!\n\r");
        return;
    }
	else
		xil_printf("EMAC Initialization Success !!\n\r");


	Result = InitializeTimerCounter();
	if (Result != XST_SUCCESS)
    {
		xil_printf("TimerCounter Initialization Error !!\n\r");
        return;
    }
	else
		xil_printf("TimerCounter Initialization Success !!\n\r");

    Result = SetupInterruptSystem(EmacPtr);
    if (Result != XST_SUCCESS)
    {
		xil_printf("Set Interrupt System Error !!\n\r");
        return;
    }
	else
		xil_printf("Set Interrupt System Success !!\n\r");


	Result = Initialize_GPio();	
    if (Result != XST_SUCCESS)
    {
		xil_printf("GPio Initialization Error !!\n\r");
        return;
    }
	else
		xil_printf("GPio Initialization Success !!\n\r");
}
