/******************************************************************************
*
*	Simple Reliable Communication System over Ethernet _ Single MicroBlaze System 
*
*	Source Code for the Send Functions	
* 
* 	Author: 	Patrick E. Akl  
*           	American University of Beirut
*				August 2004
*			 	
*	For any comments, suggestions or bug discovery, please contact me at the following
*	email address : patrickakl@hotmail.com
*					
*	or contact Christopher-John Comis (University of Toronto)
*   at the following email address:
*					comis@eecg.toronto.edu
*
*******************************************************************************/
               
#include "SingleMB_SendFunctions.h"


/*
* Fills the Ethernet Header and the Payload into TxFrame Buffer
* Source Address automatically included by PHY
*
* returns XST_SUCCESS or XST_FAILURE for just sending the frame
*
* Further reliability should be implemented seperatly
*
* Ethernet Header Format: 	Destination Address _ 6 bytes 
*							Source Address      _ 6 bytes
*                           Len / Type Field    _ 2 bytes
*/
static XStatus SendFrame(XEmac *EmacPtr, Xuint32 FramePayloadSize, Xuint8* InputBuffer,	
                         Xuint8 *DestAddress)									
{
	Xuint32 Index;
    Xuint8 *FramePtr;
    Xuint8 *AddrPtr = DestAddress; 				

    FramePtr = (Xuint8 *)TxFrame;				

    *FramePtr++ = *AddrPtr++;		// Writing destination address				
    *FramePtr++ = *AddrPtr++;
    *FramePtr++ = *AddrPtr++;
    *FramePtr++ = *AddrPtr++;
    *FramePtr++ = *AddrPtr++;
    *FramePtr++ = *AddrPtr++; 
									// Source address automatically written
									// just keep place for it
    FramePtr += XEM_MAC_ADDR_SIZE;  			
					
									// Write payload Size
    *(Xuint16 *)FramePtr = (Xuint16)FramePayloadSize;		 
    
	FramePtr += 2;								 
                                    // Write Payload     	
	for (Index = 0; Index < FramePayloadSize ; Index++)	
    {														
        *FramePtr++ = *InputBuffer++;
    }			

    return XEmac_FifoSend(EmacPtr, (Xuint8 *)TxFrame, FramePayloadSize + XEM_HDR_SIZE);
}



/*
* Pass it a pointer to the message we want to send an ACK for
*
* Destination MAC address for the ACK determined automatically
*
* NOTE: this will send an ACK only for frames that are non ACK Messages
*       The check for this is in the 'if' statement
*
* Returns XST_SUCCESS or XST_FAILURE
*
* XST_FAILURE	if we failed to send
*				or if we tried to send an ACK for an ACK
*/
XStatus Send_Ack_Message(Xuint8 *MsgBuffer)  		
{
	Xuint32 mycounter;
    Xuint8 *MessagePtr;
	Xuint8 *SourceAdd;					// will hold the ack destination address 
										// which is the received message source address
	Xuint32 RecvSeqNumber;				// Received Seq. Num will be sent back in the ACK

	Xuint8 *MesgPtr = MsgBuffer;        

	MesgPtr+=6; 						// Get destination address Pointer
	SourceAdd = MesgPtr;

	MesgPtr+=8;
	
	if(*(Xuint16 *)MesgPtr!= ACK_MESSAGE_TYPE)	// Don't send an ACK for an ACK !!
	{
		MesgPtr+=2;						

		RecvSeqNumber = *(Xuint32 *)MesgPtr;	// Get the sequence number

		// Now Write the Specific Fields Frame To Send in Pre Written ACK Frame

		MessagePtr = (Xuint8 *)AckFrame;	// AckFrame will contain the whole ACK frame

		*MessagePtr++ = *SourceAdd++;		// Copying the Destination Address
		*MessagePtr++ = *SourceAdd++;
		*MessagePtr++ = *SourceAdd++;
		*MessagePtr++ = *SourceAdd++;
		*MessagePtr++ = *SourceAdd++;
		*MessagePtr++ = *SourceAdd++; 

		// Get Past Local Address (Automatically Written)	_6 bytes
		// Get Past Ethernet Len Field						_2 bytes
		// Get Past Layer 2 Type Field						_2 bytes
		// To get to the Sequence Number Position
		MessagePtr += 10;

		// Writing Sequence Number
		*(Xuint32*)MessagePtr = RecvSeqNumber;	

		return XEmac_FifoSend(EmacPtr, (Xuint8 *)AckFrame, ACK_FRAME_SIZE);	
	}	
	return XST_FAILURE; // we tried to send an ACK for an ACK
}


/*
* Fills the Layer 2 Header and the Payload into TxMessage Buffer
* It calls the SendFrame to fill the Ethernet header and send the frame
*
* Note: the messages sent by this function can only be <= 1490 bytes in length
* returns XST_SUCCESS or XST_FAILURE for just sending the frame
*
* Further reliability should be implemented seperatly
*
*  Message Header Format: 	Type			 _ 2 bytes 
*							Sequence Number	 _ 4 bytes
*                           Message Length	 _ 4 bytes
*/
static XStatus SendMessage(Xuint32 MessagePayloadSz, Xuint8* InputBuffer,	
                         Xuint8 *DestAddress,  Xuint16 TypeofMessage)									
{

	Xuint32 CommVarsIndex;
	Xuint32 mycounter;
    Xuint8 *TemporaryPtr;
    Xuint8 *MessagePtr;
	Xuint32 FrPaySz;
	Xuint8*TempAddressPtr = DestAddress;

	CommVarsIndex = GetIndex((Xuint8 *)TempAddressPtr);



	TemporaryPtr = TemporaryBuffer;
    MessagePtr = (Xuint8 *)TxMessage;		

	// Type of Message
	*(Xuint16*)MessagePtr = TypeofMessage;		

	MessagePtr+=2;

	// Sequence Number
	*(Xuint32*)TemporaryPtr = CommVars[CommVarsIndex].SeqNum;			
	
			
	*(Xuint16*)MessagePtr = *(Xuint16 *)TemporaryPtr;
	
	MessagePtr+=2;
	TemporaryPtr+=2;

	*(Xuint16*)MessagePtr = *(Xuint16 *)TemporaryPtr;	
	
	MessagePtr+=2;
	TemporaryPtr+=2;
	
	TemporaryPtr = TemporaryBuffer;
	*(Xuint32*)TemporaryPtr = MessagePayloadSz;   // Length of Message
	
	*(Xuint16*)MessagePtr = *(Xuint16 *)TemporaryPtr;   
	MessagePtr+=2;
	TemporaryPtr+=2;

	*(Xuint16*)MessagePtr = *(Xuint16 *)TemporaryPtr;	
	MessagePtr+=2;
	TemporaryPtr+=2;

	TemporaryPtr = InputBuffer;

	if(MessagePayloadSz <=MAX_L2_PAYLOAD)
		FrPaySz = MessagePayloadSz + L2_HEADER_SIZE;   	// The Frame contains the Message header (10 bytes)
														// and the whole message itself as its payload	
	   // case of fragmented message
	else
	{
		if(TypeofMessage == FIRST_FRAGMENT_TYPE || TypeofMessage == INTERNAL_FRAGMENT_TYPE) // frame is full
			FrPaySz = L2_HEADER_SIZE + MAX_L2_PAYLOAD;  // 1490 + 10
		else if (TypeofMessage == LAST_FRAGMENT_TYPE)	// last fragment of a message
		{
			if(MessagePayloadSz % MAX_L2_PAYLOAD == 0)		// 1490 payload size of last fragment
				FrPaySz = L2_HEADER_SIZE + MAX_L2_PAYLOAD;	// 1490 + 10
			else	
				FrPaySz = L2_HEADER_SIZE + (MessagePayloadSz % MAX_L2_PAYLOAD);
		}	
	}
	
	for(mycounter = 0 ; mycounter < ( FrPaySz - L2_HEADER_SIZE ) ; mycounter++)
	{	
		*MessagePtr++ = *TemporaryPtr++;  // fill layer 2 payload		
	}

	return SendFrame(EmacPtr, FrPaySz,  TxMessage, DestAddress);	
}


/*
* Uses send message to send the payload passed through the pointer InputBuffer
* 
* tried to send it 10 times after which it returns XST_FAILURE
* returns XST_SUCCESS if an ACK is received for this message
*
* Uses different waiting times for resetting the timer.
* Once the timer expires it sets Timer_Expired = 1 
*
* Ack_Received is set to 1 by the receive handler function
*
*/
static XStatus SendReliably(Xuint32 MessagePayloadSz, Xuint8 *InputBuffer, Xuint8 *DestAddress, Xuint16 TypeofMessage)	
{	
	Xuint32 CommVarsIndex;
	Xuint8* TempAddressPtr;
	Xuint8 *MssPtr;
	Xuint8	RetransmissionCounter=0;	// we still did not make retransmissions
	
	TempAddressPtr = DestAddress;
	CommVarsIndex = GetIndex((Xuint8 *)TempAddressPtr);


	MssPtr = InputBuffer;
	CommVars[CommVarsIndex].Ack_Received = 0;
	Timer_Expired = 1;					// To enter the first time
	
	while(CommVars[CommVarsIndex].Ack_Received == 0 && RetransmissionCounter < MAX_RETRANSMIT_ATTEMPTS)  
	{
		if(Timer_Expired == 1) // The timer wrapped back or we are in for the first time
		{					
			RetransmissionCounter++;
			Timer_Expired = 0;
			SendMessage(MessagePayloadSz, MssPtr, DestAddress, TypeofMessage); 

			XTmrCtr_SetResetValue(&TimerCounter,TIMER_COUNTER_0, RstVal[RetransmissionCounter%MAX_RETRANSMIT_ATTEMPTS]);
					// Setting the reset value
			XTmrCtr_Start(&TimerCounter, TIMER_COUNTER_0);		     
					// Starting to count
			XIntc_Acknowledge(&InterruptController, INTC_INPUT_TIMER);
					// deleting to 'disable' all interrupts previously done by the running counter
			XIntc_Enable(&InterruptController, INTC_INPUT_TIMER);		
					// Now we accept Acknowledgments and Pass them to Microblaze
		}
	}
	XIntc_Disable(&InterruptController, INTC_INPUT_TIMER);
					// Don't let the interrupts constantly generated by the counter call the timer handler
	if(RetransmissionCounter == MAX_RETRANSMIT_ATTEMPTS && CommVars[CommVarsIndex].Ack_Received == 0)
		return XST_FAILURE;
	else
	{
		CommVars[CommVarsIndex].SeqNum = ( CommVars[CommVarsIndex].SeqNum + 1 ) % 0xFFFFFFFF; 	
		// in case we are not able to transmit, 
		// the sequence number stays the same
	
		return XST_SUCCESS;
	}
}


/*
* Fragments the content pointed to by input buffer and sends the fragmented message across successive
* maximum payload 1490 bytes frames.
*
* In case one of the messages failed to arrive reliably to the receiver, 
* the sequence numbers are reset so what they were before starting to send the message
*
* a very unlikely deadlock case is if the last ack that should be received by this sender failed to reach
* us 10 times. In this case, the sender would reset its sequence numbers, whereas the receiver would have the waiting sequence numbers
* after the last received message.
*
* Future Work: Deal properly with this case. Maybe send some special type of message to have like a sequence number handshake
* or agreement.
*/
static XStatus Fragment_and_Send_Reliably(Xuint32 MessagePayloadSz, Xuint8 *InputBuffer, Xuint8 *DestAddress)	
{
	Xuint32 CommVarsIndex;
	Xuint32 RememberSeqNum;			// Remembers the sequence number before starting to send for recovery in case of errors
	Xuint8 *CurrentPtr;
	Xuint32 NumberofFragmentsSent;	// Number of Fragments
	Xuint32 TempSize;				// temporary remaining payload size
	XStatus Worked;					// to know if we could send the message
	Xuint8 *TempAddressPtr;

	TempAddressPtr = DestAddress;
	CommVarsIndex = GetIndex((Xuint8 *)TempAddressPtr);


	RememberSeqNum = CommVars[CommVarsIndex].SeqNum;
	TempSize = MessagePayloadSz;	// At beginning we have the whole message to send
	NumberofFragmentsSent = 0;
	CurrentPtr = InputBuffer;

	if(MessagePayloadSz <= MAX_L2_PAYLOAD)	// can fit in one frame
	{
		Worked = SendReliably(MessagePayloadSz, CurrentPtr, DestAddress, NON_FRAGMENTED_MESSAGE_TYPE);			// this is an internal frame
		if(Worked == XST_FAILURE)
		{
			CommVars[CommVarsIndex].SeqNum = RememberSeqNum;	// Redundant Here
			xil_printf("Failed to Send the Fragmented Message !!\n\r");
			return XST_FAILURE;									// exit the function
		}
	}


	else if(MessagePayloadSz <= (3 * MAX_L2_PAYLOAD))	// 3 x 1490 bytes where 1490 
														// is the maximum message payload size
	{
		while(TempSize != 0)
		{
			if(TempSize == MessagePayloadSz)			// this is the first frame (2)
			{
				Worked = SendReliably(MessagePayloadSz, CurrentPtr, DestAddress, FIRST_FRAGMENT_TYPE);
					if(Worked == XST_FAILURE)
					{
						CommVars[CommVarsIndex].SeqNum = RememberSeqNum;
						xil_printf("Failed to Send the Fragmented Message !!\n\r");
						return XST_FAILURE;				// exit the function
					}
				TempSize		-=	MAX_L2_PAYLOAD;		// we have 1490 less bytes to send
				CurrentPtr		+=	MAX_L2_PAYLOAD;		// Increment the Current Pointer by 1490 bytes

			}
			else if(TempSize > MAX_L2_PAYLOAD)			// The Layer 2 Header is 10 bytes and 
														// the maximum Ethernet payload is 1500 bytes
			{	
				Worked = SendReliably(MessagePayloadSz, CurrentPtr, DestAddress, INTERNAL_FRAGMENT_TYPE); // this is an internal frame (3)
					if(Worked == XST_FAILURE)
					{
						CommVars[CommVarsIndex].SeqNum = RememberSeqNum;
						xil_printf("Failed to Send the Fragmented Message !!\n\r");
						return XST_FAILURE;			// exit the function
					}
				TempSize		-=	MAX_L2_PAYLOAD; // we have 1490 less bytes to send
				CurrentPtr		+=	MAX_L2_PAYLOAD; // Increment the Current Pointer by 1490 bytes
			
			}		

			else if(TempSize <= MAX_L2_PAYLOAD)		// Last Frame (4)
			{
				Worked = SendReliably(MessagePayloadSz, CurrentPtr, DestAddress, LAST_FRAGMENT_TYPE); // this is the last frame 	
				if(Worked == XST_FAILURE)
				{
					CommVars[CommVarsIndex].SeqNum = RememberSeqNum;
					xil_printf("Failed to Send the Fragmented Message !!\n\r");
					return XST_FAILURE;				// exit the function
				}
				TempSize = 0;
			}
			NumberofFragmentsSent ++;
		}
	
		return XST_SUCCESS;		
	}
	
	else
	{
		xil_printf("Invalid Length __ Maximum Message Size = 4470 bytes \n\r");
	
		return XST_FAILURE;
	}

	return XST_SUCCESS;	
}
