/******************************************************************************
*
*	Simple Network Sniffer __ Initialization Functions 
*
*	These functions will initialize the GPio, the EMAC Controller, as well as the 
*	Interrupt Controller
*
*	In Main only Initialize_Parameters() is called. it turns the led in case everything
*	Was properly initialized and returns XST_SUCCESS
*
* 	Author: 	Patrick E. Akl  
*           	American University of Beirut
*				August 2004
*			 	
*	For any comments, suggestions or bug discovery, please contact me at the following
*	email address : patrickakl@hotmail.com
*					
*	or contact Christopher-John Comis (University of Toronto)
*   at the following email address:
*					comis@eecg.toronto.edu
*
*******************************************************************************/

#include "Sniffer_GlobalVariables.h" 
#include "Sniffer_Handlers.h"

/*
* Initializes the GPio
*	DIP Switch is used to select the amount of information to sniff	
*	Led is an output
*/
static XStatus Initialize_GPio()								
{
	XStatus Status;
    Status = XGpio_Initialize(&Gpio, XPAR_OPB_GPIO_0_DEVICE_ID);
    if (Status != XST_SUCCESS)
    {
        xil_printf("GPIO initialization error !!\n\r");
    }
	else   
		xil_printf("GPIO initialization succeeded !!\n\r");
    /* Set the direction for all signals to be inputs except the LED 
     * outputs */
    XGpio_SetDataDirection(&Gpio, ~LED);

	return Status;
}


/*
* Setup Interrupt System
* EMAC has input 0 of the interrupt controller
*/

static XStatus SetupInterruptSystem(XEmac *EmacPtr)			/* Setup Interrupt System */
{
    XStatus Result;

    microblaze_enable_interrupts();

    Result = XIntc_Initialize(&InterruptController, INTC_DEVICE_ID);    
	if (Result != XST_SUCCESS)
    {
        xil_printf("Error in Xintc Initialization !!\n\r");
		return Result;
    }
	else
        xil_printf("Xintc Initialization Succeeded !!\n\r");


	Result = XIntc_Connect(&InterruptController, INTC_INPUT_EMAC, (XInterruptHandler)XEmac_IntrHandlerFifo, EmacPtr);
    if (Result != XST_SUCCESS)
    {
		return Result;
        xil_printf("Error in Xintc Connect to EMAC!!\n\r");    
    }
	else
        xil_printf("Xintc Connect Succeeded to EMAC!!\n\r");
					

	Result = XIntc_SetOptions(&InterruptController, XIN_SVC_ALL_ISRS_OPTION);  
	if (Result != XST_SUCCESS)
    {
		return Result;
        xil_printf("Error in Xintc SetOptions !!\n\r");
    }
	else
        xil_printf("Xintc SetOptions Succeeded !!\n\r");


	Result = XIntc_Start(&InterruptController, XIN_REAL_MODE);  
	if (Result != XST_SUCCESS)
    {
		return Result;
        xil_printf("Error in Xintc Start !!\n\r");
    }
	else
        xil_printf("Xintc Start Succeeded !!\n\r");
	

    XIntc_Enable(&InterruptController, INTC_INPUT_EMAC); 

    return XST_SUCCESS; 

}


/*
* Initialize the EMAC Module
* It is configured in promiscuous mode 
* for network monitoring
*/
static XStatus InitializeEmac()
{
	XEmac_Config *ConfigPtr;	
	XStatus Result;
	Xuint32 DeviceId;
	DeviceId = 0;																	
	EmacPtr = & Emac;

    ConfigPtr = XEmac_LookupConfig(DeviceId);        
    ConfigPtr->IpIfDmaConfig = XEM_CFG_NO_DMA;

    Result = XEmac_Initialize(EmacPtr, DeviceId);
    if (Result != XST_SUCCESS)
    {	
		xil_printf("Error in EMAC Initialization !!\n\r");
        return Result;
    }
	else
		xil_printf("Initialization Succeeded !!\n\r");


    Result = XEmac_SelfTest(EmacPtr);
    if (Result != XST_SUCCESS)
    {
		xil_printf("Self Test Error !!\n\r");
        return Result;
    }
	else
		xil_printf("Self Test Succeeded !!\n\r");

    Result = XEmac_SetOptions(EmacPtr, TEST_1_OPTIONS);
    if (Result != XST_SUCCESS)
    {
		xil_printf("Set Options Error !!\n\r");
        return Result;
    }
	else
		xil_printf("Set Options Succeeded !!\n\r");

    Result = XEmac_SetMacAddress(EmacPtr, LocalAddress); 
    if (Result != XST_SUCCESS)
    {
		xil_printf("Set MAC Address Error !!\n\r");
        return Result;
    }
	else
		xil_printf("Set MAC Address Succeeded !!\n\r");

    XEmac_SetFifoSendHandler(EmacPtr, EmacPtr, FifoSendHandler);
    XEmac_SetFifoRecvHandler(EmacPtr, EmacPtr, FifoRecvHandler);
    XEmac_SetErrorHandler(EmacPtr, EmacPtr, ErrorHandler);


	XEmac_ClearStats(EmacPtr);
    Result = XEmac_Start(EmacPtr);
    if (Result != XST_SUCCESS)
    {
		xil_printf("EMAC Start Error !!\n\r");
        return Result;
    }
	else
		xil_printf("EMAC Start Success !!\n\r");

	return XST_SUCCESS;
}


/*
* This function should be called from main
* It calls all the other functions.
*/
void Initialize_Parameters()					
{	
	XStatus Result;

	xil_printf("Initializing Sniffer\n\r");
	xil_printf("____________________\n\n\r");

	DegreeofInformation = 1; // more information

	Result = InitializeEmac();
	if (Result != XST_SUCCESS)
    {
		xil_printf("EMAC Initialization Error !!\n\r");
        return;
    }
	else
		xil_printf("EMAC Initialization Success !!\n\r");


    Result = SetupInterruptSystem(EmacPtr);
    if (Result != XST_SUCCESS)
    {
		xil_printf("Set Interrupt System Error !!\n\r");
        return;
    }
	else
		xil_printf("Set Interrupt System Success !!\n\r");


	Result = Initialize_GPio();	
    if (Result != XST_SUCCESS)
    {
		xil_printf("GPio Initialization Error !!\n\r");
        return;
    }
	else
		xil_printf("GPio Initialization Success !!\n\r");
}


